//*****************************************************************************
// Description     : Main sample program of list mode
// Target Device   : APV8508, APV8104
// Target Platform : Linux, C
// Date            : January 2025
// Version         : 2.0
//*****************************************************************************
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>

#include "sub_sitcp.h"

//
// Modify the following defines according to your environment
//
#define DEVICE_IPADDR     "192.168.10.128"  // IP address of target device
#define LIST_FILENAME    "./list_data.bin"  // name of the file where list data saved
#define MEASUREMENT_TIME_MSEC       (3000)  // measurement time (msec)

#define IPADDR_OWNPC             "0.0.0.0"  // IP address of own PC; Normally leave 0.0.0.0
#define DEVICE_PORT_CTRL_OWNPC     (50000)  // source port number for command
#define DEVICE_SOCKBUF_CMD     (16 * 1024)  // for command communication(16kB)
#define DEVICE_SOCKBUF_DATA  (1024 * 1024)  // for receiving data(1MB)
#define LIST_TRANSFER_LEN          (16000)  // length of list data per transfer

//
// Specific to device
//
#define DEVICE_PORT_DATARECV    (24)  // port number for data reception
#define DEVICE_CNT_PER_NSEC     (8)  // 1count = 8nsec


//
// Static global variables
//
static int cmd_socket;	// socket for command communication
static int data_socket;	// socket for data transfer

//#########################################################################
//  Name      : main
//  Function  : entry point
//  Arguments :
//      in argc : number of parameters
//      in argv : parameters
//  Return :
//      0 : success
//      non-0 : failure
//  Note : 
//#########################################################################
int main(int argc, char **argv)
{
	int err_cnt;
	int ret;
	unsigned long long mtm_count;
	struct sockaddr_in saddr;
	int optval;
	unsigned long long now_count;
	unsigned int now_msec;
	unsigned short status;
	unsigned int throughput_cps;
	FILE *fp;
	char list_data[LIST_TRANSFER_LEN];
	int recv_cnt, total_list_bytes;

	// convert msec to count
	mtm_count = (((unsigned long long)(MEASUREMENT_TIME_MSEC) * 1000000UL)
				/ (unsigned long long)(DEVICE_CNT_PER_NSEC));

	// ===========================================================================
	// prepare UDP socket for command communication
	cmd_socket = socket(AF_INET, SOCK_DGRAM, 0);
	if (cmd_socket > 0) { // socket created
		// change maximum socket receive buffer size
		optval = DEVICE_SOCKBUF_CMD;
		setsockopt(cmd_socket, SOL_SOCKET, SO_RCVBUF, &optval, sizeof(optval));

		// assign own address to the socket
		memset(&saddr, 0, sizeof(saddr));
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(DEVICE_PORT_CTRL_OWNPC);  // port number for receiving cmd reply
		saddr.sin_addr.s_addr = inet_addr(IPADDR_OWNPC); // IP address of own PC
		ret = bind(cmd_socket, (struct sockaddr*)(&saddr), sizeof(saddr));
		if (ret != 0) {
			printf("bind error(%d)\n", errno);
			close(cmd_socket);
		}
	}
	else {
		printf("create UDP socket error(%d)\n", errno);
		return(1);
	}

	// ===========================================================================
	// prepare TCP socket for data transfer
	data_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (data_socket > 0) { // socket created
		// change maximum socket receive buffer size
		optval = DEVICE_SOCKBUF_DATA;
		setsockopt(data_socket, SOL_SOCKET, SO_RCVBUF, &optval, sizeof(optval));

		// connect to target device
		memset(&saddr, 0, sizeof(saddr));
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(DEVICE_PORT_DATARECV);     // port number for receiving data
		saddr.sin_addr.s_addr = inet_addr(DEVICE_IPADDR); // IP address of target device
		ret = connect(data_socket, (struct sockaddr*)(&saddr), sizeof(saddr));
		if (ret != 0) {
			printf("connect error(%d)\n", errno);
			close(data_socket);

			close(cmd_socket);
		}
	}
	else {
		printf("create TCP socket error(%d)\n", errno);
		close(cmd_socket);
		return(2);
	}

	// ===========================================================================
	// set parameters to the device

	err_cnt = 0;
	
	// mode: list
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000000, 2);
	if (ret < 0) err_cnt++;

	// measurement time
	ret = sub_sitcp_write_8byte(DEVICE_IPADDR, cmd_socket, 0xB4000006, mtm_count);
	if (ret < 0) err_cnt++;

	// Trigger Select: CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000030, 1);
	if (ret < 0) err_cnt++;

	// Edge Select: 255
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000032, 255);
	if (ret < 0) err_cnt++;

	// Wave Trigger Point: 17
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400008C, 17);
	if (ret < 0) err_cnt++;

	// polarity: Positive CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400011A, 0);
	if (ret < 0) err_cnt++;

	// polarity: Positive CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400021A, 0);
	if (ret < 0) err_cnt++;

	// Free Run: OFF
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000036, 0);
	if (ret < 0) err_cnt++;

	// QDC Full Scale: 1/16 Ch1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400010C, 4);
	if (ret < 0) err_cnt++;

	// QDC Full Scale: 1/16 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400020C, 4);
	if (ret < 0) err_cnt++;

	// CFD Function: 0.21 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000160, 7);
	if (ret < 0) err_cnt++;

	// CFD Function: 0.21 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000260, 7);
	if (ret < 0) err_cnt++;

	// CFD Delay: 10ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000162, 9);
	if (ret < 0) err_cnt++;

	// CFD Delay: 10ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000262, 9);
	if (ret < 0) err_cnt++;

	// CFD Walk CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000164, 20);
	if (ret < 0) err_cnt++;

	// CFD Walk CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000264, 20);
	if (ret < 0) err_cnt++;

	// CFD Threshold: 50 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000166, 50);
	if (ret < 0) err_cnt++;

	// CFD Threshold: 50 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000266, 50);
	if (ret < 0) err_cnt++;

	// CFD LLD: 20 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000168, 10);
	if (ret < 0) err_cnt++;

	// CFD LLD: 20 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000268, 10);
	if (ret < 0) err_cnt++;

	// CFD ULD: 8190 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400016A, 8190);
	if (ret < 0) err_cnt++;

	// CFD ULD: 8190 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400026A, 8190);
	if (ret < 0) err_cnt++;

	// Baseline Restorer Filter: 128 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400016E, 128);
	if (ret < 0) err_cnt++;

	// Baseline Restorer Filter: 128 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400026E, 128);
	if (ret < 0) err_cnt++;

	// QDC Pretrigger: 8ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001C0, 1);
	if (ret < 0) err_cnt++;

	// QDC Pretrigger: 8ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002C0, 1);
	if (ret < 0) err_cnt++;

	// Wave Select: 0 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001C2, 0);
	if (ret < 0) err_cnt++;

	// Wave Select: 0 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002C2, 256);
	if (ret < 0) err_cnt++;

	// QDC Filter: 10ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001C6, 1);
	if (ret < 0) err_cnt++;

	// QDC Filter: 10ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002C6, 1);
	if (ret < 0) err_cnt++;

	// QDC Select: Sum CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001C8, 1);
	if (ret < 0) err_cnt++;

	// QDC Select: Sum CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002C8, 1);
	if (ret < 0) err_cnt++;

	// Analog Gain: x3 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400010E, 0);
	if (ret < 0) err_cnt++;

	// Analog Gain: x3 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400020E, 0);
	if (ret < 0) err_cnt++;

	// Analog Offset: 2048 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000170, 2048);
	if (ret < 0) err_cnt++;

	// Analog Offset: 2048 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000270, 2048);
	if (ret < 0) err_cnt++;

	// ADC Param. Write Enable: Enable CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001B0, 1);
	if (ret < 0) err_cnt++;

	// ADC Param. Write Enable: Enable CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002B0, 1);
	if (ret < 0) err_cnt++;

	// ADC Coarse Gain 0: 14 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001B8, 14);
	if (ret < 0) err_cnt++;

	// ADC Coarse Gain 0: 14 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002B8, 14);
	if (ret < 0) err_cnt++;

	// ADC Coarse Gain 1: 14 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001BA, 14);
	if (ret < 0) err_cnt++;

	// ADC Coarse Gain 1: 14 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002BA, 14);
	if (ret < 0) err_cnt++;

	// FRC Start Count: 8 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001D8, 8);
	if (ret < 0) err_cnt++;

	// FRC Start Count: 8 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002D8, 8);
	if (ret < 0) err_cnt++;

	// FRC Stop Count: 5 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001DA, 5);
	if (ret < 0) err_cnt++;

	// FRC Stop Count: 5 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002DA, 5);
	if (ret < 0) err_cnt++;

	// QDC Integral range : 18(=144ns/8digit) CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001DC, 18);
	if (ret < 0) err_cnt++;

	// QDC Integral range : 18(=144ns/8digit) CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002DC, 18);
	if (ret < 0) err_cnt++;

	// Input Signal Type: 0 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001DE, 0);
	if (ret < 0) err_cnt++;

	// Input Signal Type: 0 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002DE, 0);
	if (ret < 0) err_cnt++;

	// Rise Judge Threshold: 200 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000110, 200);
	if (ret < 0) err_cnt++;

	// Rise Judge Threshold: 200 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000210, 200);
	if (ret < 0) err_cnt++;

	// Timestampe Timing: CFD CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001D0, 0);
	if (ret < 0) err_cnt++;
	
	// Timestampe Timing: CFD CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002D0, 0);
	if (ret < 0) err_cnt++;

	// clear data
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000090, 0);
	if (ret < 0) err_cnt++;
	
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000090, 1);
	if (ret < 0) err_cnt++;
	
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000090, 0);
	if (ret < 0) err_cnt++;

	// wait for completion of clear processing
	usleep(200*1000);

	// measurement mode: real time
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000002, 0);
	if (ret < 0) err_cnt++;

	// List Wave Delay : 10(10digit*8ns=80ns Delay) CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000174, 10);
	if (ret < 0) err_cnt++;

	// List Wave Delay : 10(10digit*8ns=80ns Delay) CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000274, 10);
	if (ret < 0) err_cnt++;

	// List Pileup : 15 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000172, 15);
	if (ret < 0) err_cnt++;

	// List Pileup : 15 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000272, 15);
	if (ret < 0) err_cnt++;

	// ADC Sampling Bit: 14bits
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000026, 0);
	if (ret < 0) err_cnt++;

	// Signal Delay : 0(0digit*8ns=0ns Delay) CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000176, 0);
	if (ret < 0) err_cnt++;

	// Signal Delay : 0(0digit*8ns=0ns Delay) CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000276, 0);
	if (ret < 0) err_cnt++;

	// List Wave Data : 125 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400017A, 125);
	if (ret < 0) err_cnt++;

	// List Wave Data : 125 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400027A, 125);
	if (ret < 0) err_cnt++;

	// Or Output Length Setting(Option) : 25
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000070, 25);
	if (ret < 0) err_cnt++;

	//  Or Output ON/OFF 1bit(Option) : 0 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000180, 0);
	if (ret < 0) err_cnt++;

	//  Or Output ON/OFF 1bit(Option) : 0 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000280, 0);
	if (ret < 0) err_cnt++;

	//  Write wait num 4bit : 8
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400004A, 8);
	if (ret < 0) err_cnt++;
	
	//  Wave Compression Settings 4bit: 0
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400008A, 0);
	if (ret < 0) err_cnt++;

	//  Rise Start Cnt: 5ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001E8, 5);
	if (ret < 0) err_cnt++;

	//  Rise Start Cnt: 5ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002E8, 5);
	if (ret < 0) err_cnt++;

	//  Rise Stop Cnt: 10ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001EA, 10);
	if (ret < 0) err_cnt++;

	//  Rise Stop Cnt: 10ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002EA, 10);
	if (ret < 0) err_cnt++;

	//  Total Start Cnt: 5ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001EC, 5);
	if (ret < 0) err_cnt++;

	//  Total Start Cnt: 5ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002EC, 5);
	if (ret < 0) err_cnt++;

	//  Total Stop Cnt: 20ns CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001EE, 20);
	if (ret < 0) err_cnt++;

	//  Total Stop Cnt: 20ns CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002EE, 20);
	if (ret < 0) err_cnt++;

	//  PSA full scalet : 1/4 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40001D6, 2);
	if (ret < 0) err_cnt++;

	//  PSA full scalet : 1/4 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB40002D6, 2);
	if (ret < 0) err_cnt++;

	// Pileup Reject Setting(Option) 15bit: 5 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000186, 5);
	if (ret < 0) err_cnt++;

	// Pileup Reject Setting(Option) 15bit: 5 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000286, 5);
	if (ret < 0) err_cnt++;

	// List Wave Sampling  : 0(1ns) CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400018A, 0);
	if (ret < 0) err_cnt++;

	// List Wave Sampling  : 0(1ns) CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400028A, 0);
	if (ret < 0) err_cnt++;

	// Fixed Baseline Setting : 8192 CH1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400018C, 8192);
	if (ret < 0) err_cnt++;
	
	// Fixed Baseline Setting : 8192 CH2
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB400028C, 8192);
	if (ret < 0) err_cnt++;
	
	// PSD ON/OFF
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000080, 1);
	if (ret < 0) err_cnt++;

	// Board Calibration : 0
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000140, 0);
	if (ret < 0) err_cnt++;

	// Board Calibration : 1
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000140, 1);
	if (ret < 0) err_cnt++;

	// Board Calibration : 0
	ret = sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000140, 0);
	if (ret < 0) err_cnt++;

	// wait for completion of device settings
	usleep(100*1000);

	if (err_cnt > 0) {
		printf("config error(%d)...\n", err_cnt);
		close(cmd_socket);
		close(data_socket);
		return(3);
	}

	// ===========================================================================
	// open file in advance
	fp = fopen(LIST_FILENAME, "wb");
	if (fp == NULL) {
		printf("Failed to create file(%s)\n", LIST_FILENAME);
		close(cmd_socket);
		close(data_socket);
		return(4);
	}

	// ===========================================================================
	// start measurement
	// clear data
	sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000090, 0);
	sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000090, 1);
	sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000090, 0);

	// wait for completion of clear processing
	usleep(200*1000);
	// stop once, then start
	sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000004, 1);

	// read measurement status: 2byte value
	sub_sitcp_read_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000004, &ret);

	int ret_dmy = 0;
	do
	{
		// receive and write list data to file
		memset(list_data, 0, sizeof(list_data));
		ret_dmy = recv(data_socket, list_data, LIST_TRANSFER_LEN, MSG_DONTWAIT);
		if (ret_dmy > 0) {
			fwrite(list_data, (size_t)1, (size_t)ret_dmy, fp);
			total_list_bytes += ret_dmy;
		}
	} while(ret_dmy > 0);
	usleep(2000*1000);
	
	printf("start measurement!!!\n");

	// ===========================================================================
	// main measuremet loop
	recv_cnt = 0;
	total_list_bytes = 0;
	while(1) {
		if (recv_cnt % 500 == 0) { // read status every 500 receiving list
			// read measurement status: 2byte value
   			status = sub_sitcp_read_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000004, &ret);
   			if (ret != 0) break;

			// read throughput count rate (CH1 only): 4byte value
			throughput_cps = sub_sitcp_read_4byte(DEVICE_IPADDR, cmd_socket, 0xB4000130, &ret); 
	   		if (ret != 0) break;

			// read elapsed time: 8byte value
			now_count = sub_sitcp_read_8byte(DEVICE_IPADDR, cmd_socket, 0xB400000E, &ret);
			if (ret != 0) break;
			// convert count to msec
			now_msec = (unsigned int)((now_count * (unsigned long long)(DEVICE_CNT_PER_NSEC)) / 1000000UL);

			// display status
			printf("time(msec): %5d/%5d.", now_msec, (unsigned int)(MEASUREMENT_TIME_MSEC));
			printf("   throughput rate(cps): %5d\n", throughput_cps);

			// break a loop when status is 'stopped' or reached to measurement time
			if ((status == 0) || (now_count >= mtm_count)) {
				break;
			}
		}

		// receive and write list data to file
		memset(list_data, 0, sizeof(list_data));
		ret = recv(data_socket, list_data, LIST_TRANSFER_LEN, MSG_DONTWAIT);
		if (ret > 0) {
			fwrite(list_data, (size_t)1, (size_t)ret, fp);
			total_list_bytes += ret;
		}
		recv_cnt++;

		// wait a while
		usleep(1000);
	} // end of measurement loop

	// close file
	fclose(fp);

	if (ret >= 0) {
		printf("measurement completed!!!\n");
	}
	else {
		printf("measurement error(%d)...\n", ret);
	}

	if (total_list_bytes > 0) printf("total list size = %d bytes\n", total_list_bytes);

	// stop measurement
	sub_sitcp_write_2byte(DEVICE_IPADDR, cmd_socket, 0xB4000004, 0);

	// close sockets
	close(cmd_socket);
	close(data_socket);
	return(0);
}
