//*****************************************************************************
// Description     : Definition of functions for SiTCP communication
// Target Device   : any device with SiTCP board
// Target Platform : Linux, C
// Date            : January 2025
// Version         : 2.0
//*****************************************************************************
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "sub_sitcp.h"


//******************** macro definition ********************
// RBCP packet: version, type
#define RBCP_VER_TYPE	((unsigned char)(0xff))

// RBCP packet: read command flag
#define RBCP_CMDFLAG_READ_REQ	((unsigned char)(0xc0))	// request
#define RBCP_CMDFLAG_READ_ACK	((unsigned char)(0xc8))	// acknowledge

// RBCP packet: write command flag
#define RBCP_CMDFLAG_WRITE_REQ	((unsigned char)(0x80))	// request
#define RBCP_CMDFLAG_WRITE_ACK	((unsigned char)(0x88))	// acknowledge

// RBCP packet: identifier(any)
#define RBCP_ID	(7)

// RBCP packet: data length
#define RBCP_DATA_LEN	(2)

// RBCP packet length
#define RBCP_PACKET_LEN_ADDR_ONLY	(8)		// address only
		// header(=4) + address(=4)
#define RBCP_PACKET_LEN_WITH_DATA	(10)	// address and data
		// header(=4) + address(=4) + data(=2)


// destination port number for command
#define DEVICE_PORT_CTRL_DEVICE   (4660)

// timeout value for waiting command response
#define DEVICE_TIMEOUT_CMD	(200)


//******************** structure type definition ********************
// RBCP packet header
typedef struct rbcp_header {
	unsigned char ver_type;	// version(4bit), type(4bit)
	unsigned char cmd_flag;	// command(4bit), flag(4bit)
	unsigned char id;		// identifier
	unsigned char length;	// data length(byte)
} RBCP_HEADER;

// RBCP packet format for address only (for Status command)
typedef struct rbcp_packet_addr_only {
	RBCP_HEADER header;
	unsigned int address;	// command address(32bit)
} RBCP_PACKE_ADDR_ONLY;

// RBCP packet format for address and data
typedef struct rbcp_packet_with_data {
	RBCP_HEADER header;
	unsigned int address;	// command address(32bit)
	unsigned short data;	// data(16bit)
} RBCP_PACKE_WITH_DATA;


//
// Local functions declaration
//
static int sitcp_read(char *, int, unsigned int, unsigned short *);
static int sitcp_write(char *, int, unsigned int, unsigned short);
static int sitcp_recv_cmd(int, char *, int, int);


//#########################################################################
//  Name      : sitcp_read
//  Function  : command communication(read request)
//  Arguments :
//      in  pIpAddr  : target IP address
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      out pData    : pointer to the response data
//  Return :
//       0 : success
//      -1 : no response within a period of timeout value
//      -2 : invalid size of received packet
//      -3 : received packet is not a response to the request
//#########################################################################
int sitcp_read(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, unsigned short *pData)
{
	struct sockaddr_in saddr;
	RBCP_PACKE_ADDR_ONLY reqdt;
	RBCP_PACKE_WITH_DATA rcvdt;
	RBCP_HEADER cmpdt;
	int result;
	int sbyte, rbyte;
	int ssize, rsize;

	result = 0;

	// clear area for reception
	memset(&rcvdt, 0, sizeof(rcvdt));

	// set size of packet for request and response
	ssize = RBCP_PACKET_LEN_ADDR_ONLY;
	rsize = RBCP_PACKET_LEN_WITH_DATA;

	// construct packet for read request
	memset(&reqdt, 0, sizeof(reqdt));
	reqdt.header.ver_type = RBCP_VER_TYPE;
	reqdt.header.cmd_flag = RBCP_CMDFLAG_READ_REQ;
	reqdt.header.id = RBCP_ID;
	reqdt.header.length = RBCP_DATA_LEN;
	reqdt.address = htonl(cmd_addr);

	// set target address information
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(DEVICE_PORT_CTRL_DEVICE);	// target port
	saddr.sin_addr.s_addr = inet_addr(pIpAddr);	// target IP address

	// send the packet to the target
	sbyte = sendto(sock_cmd, (char*)(&reqdt), ssize, 0,
				  (struct sockaddr*)(&saddr), sizeof(saddr));

	if (sbyte > 0) {
		// receive response
		rbyte = sitcp_recv_cmd(sock_cmd, (char*)(&rcvdt), rsize, DEVICE_TIMEOUT_CMD);

		if (rbyte >= rsize) {
			// construct header part of response packet from request one
			// by change flag, which is only difference, from 'request' to 'acknowledge'
			cmpdt = reqdt.header;
			cmpdt.cmd_flag = RBCP_CMDFLAG_READ_ACK;

			// check if header and address is expected one
			if ( (memcmp(&cmpdt, &rcvdt.header, sizeof(cmpdt)) != 0) || (reqdt.address != rcvdt.address) ) {
				result = -3;	// not a response to the request
			}
		}
		else { // short of data size
			result = -2;
		}
	}
	else { // receive error
		result = -1;
	}

	// pass response data to caller
	*pData = htons(rcvdt.data);

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_read_2byte
//  Function  : command communication(read request for 2byte value)
//  Arguments :
//      in  pIpAddr  : target IP address
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//#########################################################################
unsigned short sub_sitcp_read_2byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, int *errcd)
{
	int ret;
	unsigned short result = 0;

	ret = sitcp_read(pIpAddr, sock_cmd, cmd_addr, &result);

	if (errcd != NULL) {
		*errcd = ret;
	}

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_read_4byte
//  Function  : command communication(read request for 4byte value)
//  Arguments :
//      in  pIpAddr  : target IP address
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//#########################################################################
unsigned int sub_sitcp_read_4byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, int *errcd)
{
	int i, ret;
	unsigned short rdat[2];
	unsigned int result = 0;

	memset(&rdat, 0, sizeof(rdat));

	// read data in each 2bytes (start with MSB)
	for (i = 0; i < 2; i++) {
		ret = sitcp_read(pIpAddr, sock_cmd, cmd_addr + i*2, &rdat[i]);
		if (ret == 0) {
			result += (unsigned int)( (unsigned int)rdat[i] << (1-i)*16 );
		} else {
			break;
		}
	}

	if (errcd != NULL) {
		*errcd = ret;
	}

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_read_6byte
//  Function  : command communication(read request for 6byte value)
//  Arguments :
//      in  pIpAddr  : target IP address
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//#########################################################################
unsigned long long sub_sitcp_read_6byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, int *errcd)
{
	int i, ret;
	unsigned short rdat[3];
	unsigned long long result = 0;

	memset(&rdat, 0, sizeof(rdat));

	// read data in each 2bytes (start with MSB)
	for (i = 0; i < 3; i++) {
		ret = sitcp_read(pIpAddr, sock_cmd, cmd_addr + i*2, &rdat[i]);
		if (ret == 0) {
			result += (unsigned long long)( (unsigned long long)rdat[i] << (2-i)*16 );
		} else {
			break;
		}
	}

	if (errcd != NULL) {
		*errcd = ret;
	}

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_read_8byte
//  Function  : command communication(read request for 8byte value)
//  Arguments :
//      in  pIpAddr  : target IP address
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//#########################################################################
unsigned long long sub_sitcp_read_8byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, int *errcd)
{
	int i, ret;
	unsigned short rdat[4];
	unsigned long long result = 0;

	memset(&rdat, 0, sizeof(rdat));

	// read data in each 2bytes (start with MSB)
	for (i = 0; i < 4; i++) {
		ret = sitcp_read(pIpAddr, sock_cmd, cmd_addr + i*2, &rdat[i]);
		if (ret == 0) {
			result += (unsigned long long)( (unsigned long long)rdat[i] << (3-i)*16 );
		} else {
			break;
		}
	}

	if (errcd != NULL) {
		*errcd = ret;
	}

	return(result);
}


//#########################################################################
//  Name      : sitcp_write
//  Function  : command communication(write request)
//  Arguments :
//      in  pIpAddr  : target IP address
//      in  sock_cmd : file descriptor for command communication
//      in  cmd_addr : command address
//      in  cmd_data : data to be written
//  Return :
//       0 : success
//      -1 : no response within a period of timeout value
//      -2 : invalid size of received packet
//      -3 : received packet is not a response to the request
//#########################################################################
int sitcp_write(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, unsigned short cmd_data)
{
	struct sockaddr_in saddr;
	RBCP_PACKE_WITH_DATA reqdt, rcvdt;
	RBCP_HEADER cmpdt;
	int result;
	int sbyte, rbyte;
	int ssize, rsize;

	result = 0;

	// clear area for reception
	memset(&rcvdt, 0, sizeof(rcvdt));

	// set size of packet for request and response
	ssize = RBCP_PACKET_LEN_WITH_DATA;
	rsize = ssize;

	// construct packet for write request
	reqdt.header.ver_type = RBCP_VER_TYPE;
	reqdt.header.cmd_flag = RBCP_CMDFLAG_WRITE_REQ;
	reqdt.header.id = RBCP_ID;
	reqdt.header.length = RBCP_DATA_LEN;
	reqdt.address = htonl(cmd_addr);
	reqdt.data = htons(cmd_data);

	// set target address information
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(DEVICE_PORT_CTRL_DEVICE);	// target port
	saddr.sin_addr.s_addr = inet_addr(pIpAddr);	// target IP address

	// send the packet to the target
	sbyte = sendto(sock_cmd, (char*)(&reqdt), ssize, 0,
				  (struct sockaddr*)(&saddr), sizeof(saddr));

	if (sbyte > 0) {
		// receive response
		rbyte = sitcp_recv_cmd(sock_cmd, (char*)(&rcvdt), rsize, DEVICE_TIMEOUT_CMD);

		if (rbyte >= rsize) {
			// construct header part of response packet from request one
			// by change flag, which is only difference, from 'request' to 'acknowledge'
			cmpdt = reqdt.header;
			cmpdt.cmd_flag = RBCP_CMDFLAG_WRITE_ACK;

			// check if header and address is expected one
			if ( (memcmp(&cmpdt, &rcvdt.header, sizeof(cmpdt)) != 0) || (reqdt.address != rcvdt.address) ){
				result = -3;	// not a response to the request
			}
		}
		else { // short of data size
			result = -2;
		}
	}
	else { // receive error
		result = -1;
	}

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_write_2byte
//  Function  : command communication(write request for 2byte value)
//  Arguments :
//      in pIpAddr  : target IP address
//      in sock_cmd : file descriptor for command communication
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//#########################################################################
int sub_sitcp_write_2byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, unsigned short wdata)
{
	int result;

	result = sitcp_write(pIpAddr, sock_cmd, cmd_addr, wdata);
	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_write_4byte
//  Function  : command communication(write request for 4byte value)
//  Arguments :
//      in pIpAddr  : target IP address
//      in sock_cmd : file descriptor for command communication
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//#########################################################################
int sub_sitcp_write_4byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, unsigned int wdata)
{
	int i, result;
	unsigned short data;

	// write data in each 2bytes (start with MSB)
	for(i = 0; i < 2; i++) {
		data = (unsigned short)( 0xFFFF & (wdata >> ((1-i)*16)) );
		result = sitcp_write(pIpAddr, sock_cmd, cmd_addr + i*2, data);
		if (result < 0) break;
	}

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_write_6byte
//  Function  : command communication(write request for 6byte value)
//  Arguments :
//      in pIpAddr  : target IP address
//      in sock_cmd : file descriptor for command communication
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//#########################################################################
int sub_sitcp_write_6byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, unsigned long long wdata)
{
	int i, result;
	unsigned short data;

	// write data in each 2bytes (start with MSB)
	for(i = 0; i < 3; i++) {
		data = (unsigned short)( 0xFFFF & (wdata >> ((2-i)*16)) );
		result = sitcp_write(pIpAddr, sock_cmd, cmd_addr + i*2, data);
		if (result < 0) break;
	}

	return(result);
}


//#########################################################################
//  Name      : sub_sitcp_write_8byte
//  Function  : command communication(write request for 8byte value)
//  Arguments :
//      in pIpAddr  : target IP address
//      in sock_cmd : file descriptor for command communication
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//#########################################################################
int sub_sitcp_write_8byte(char *pIpAddr, int sock_cmd, unsigned int cmd_addr, unsigned long long wdata)
{
	int i, result;
	unsigned short data;

	// write data in each 2bytes (start with MSB)
	for(i = 0; i < 4; i++) {
		data = (unsigned short)( 0xFFFF & (wdata >> ((3-i)*16)) );
		result = sitcp_write(pIpAddr, sock_cmd, cmd_addr + i*2, data);
		if (result < 0) break;
	}

	return(result);
}


//#########################################################################
//  Name      : sitcp_recv_cmd
//  Function  : command communication(receive response)
//  Arguments :
//      in sock    : file descriptor for command communication
//      in rbuffer : buffer for received data
//      in size    : size of data to be received(byte)
//      in timeout : receive timeout value(msec)
//  Return :
//      positive value: received data size
//      -11 : no response within a period of timeout value
//      -12 : no data
//#########################################################################
int sitcp_recv_cmd(int sock, char *rbuffer, int size, int timeout)
{
	struct sockaddr_in saddr;
	struct timeval tval;
	socklen_t sock_len;
	fd_set fdsock;
	int ret;
	int result;
	int recv_size;

	result = 0;

	// set timeout value
	memset(&tval, 0, sizeof(tval));
	tval.tv_sec = (int)(timeout / 1000);
	tval.tv_usec = (int)((timeout % 1000) * 1000);

	// wait until received data is ready
	FD_ZERO(&fdsock);
	FD_SET(sock, &fdsock);
	ret = select((sock + 1), &fdsock, NULL, NULL, &tval);

	if (ret != 0) {
		if (FD_ISSET(sock, &fdsock) != 0) {
			// receive message
			sock_len = sizeof(saddr);
			memset(&saddr, 0, sizeof(saddr));
			recv_size = recvfrom(sock, rbuffer, size, 0, (struct sockaddr*)(&saddr), &sock_len);
			result = (int)(recv_size);	// set received data size
		}
		else { // no data
			result = -12;
		}
	}
	else { // timeout
		result = -11;
	}

	return(result);
}
